# ruff: noqa: D100, D101, D102, D103, D104, D107, N999
from __future__ import annotations

from enum import StrEnum

from redux import BaseAction, Immutable


class SoundDevice(StrEnum):
    INPUT = 'Input'
    OUTPUT = 'Output'


class SoundSetVolumeAction(BaseAction):
    volume: float
    device: SoundDevice


class SoundChangeVolumeAction(BaseAction):
    amount: float
    device: SoundDevice


class SoundSetMuteStatusAction(BaseAction):
    mute: bool
    device: SoundDevice


class SoundToggleMuteStatusAction(BaseAction):
    device: SoundDevice


SoundAction = (
    SoundSetVolumeAction
    | SoundChangeVolumeAction
    | SoundSetMuteStatusAction
    | SoundToggleMuteStatusAction
)


class SoundState(Immutable):
    output_volume: float
    is_output_mute: bool
    mic_volume: float
    is_mic_mute: bool



# ruff: noqa: D100, D101, D102, D103, D104, D107
from __future__ import annotations

from dataclasses import replace
from typing import Sequence

from redux import BaseAction, Immutable, InitAction, InitializationActionError


class IconState(Immutable):
    symbol: str
    color: str
    priority: int
    id: str | None


class StatusIconsState(Immutable):
    icons: Sequence[IconState]


class StatusIconsRegisterAction(BaseAction):
    icon: str
    color: str = 'white'
    priority: int = 0
    id: str | None = None


IconAction = StatusIconsRegisterAction | InitAction


def reducer(state: StatusIconsState | None, action: IconAction) -> StatusIconsState:
    if state is None:
        if isinstance(action, InitAction):
            return StatusIconsState(icons=[])
        raise InitializationActionError(action)
    if isinstance(action, StatusIconsRegisterAction):
        return replace(
            state,
            icons=sorted(
                [
                    *[
                        icon_status
                        for icon_status in state.icons
                        if icon_status.id != action.id or icon_status.id is None
                    ],
                    IconState(
                        symbol=action.icon,
                        color=action.color,
                        priority=action.priority,
                        id=action.id,
                    ),
                ],
                key=lambda entry: entry.priority,
            ),
        )
    return state



# ruff: noqa: D100, D101, D102, D103, D104, D107, N999
from pathlib import Path

from reducer import reducer
from setup import init_service

from ubo_app.load_services import register_service

register_service(
    service_id='camera',
    label='Camera',
    reducer=reducer,
)

IS_RPI = Path('/etc/rpi-issue').exists()
if IS_RPI:
    init_service()



# ruff: noqa: D100, D101, D102, D103, D104, D107, N999
from __future__ import annotations

from dataclasses import replace

from constants import SOUND_MIC_STATE_ICON_ID, SOUND_MIC_STATE_ICON_PRIORITY
from redux import (
    BaseEvent,
    CompleteReducerResult,
    InitAction,
    InitializationActionError,
    ReducerResult,
)

from ubo_app.store.sound import (
    SoundAction,
    SoundChangeVolumeAction,
    SoundDevice,
    SoundSetMuteStatusAction,
    SoundSetVolumeAction,
    SoundState,
    SoundToggleMuteStatusAction,
)
from ubo_app.store.status_icons import (
    StatusIconsRegisterAction,
)

Action = InitAction | SoundAction | StatusIconsRegisterAction


def reducer(  # noqa: C901
    state: SoundState | None,
    action: Action,
) -> ReducerResult[SoundState, Action, BaseEvent]:
    if state is None:
        if isinstance(action, InitAction):
            return SoundState(
                output_volume=0.5,
                is_output_mute=False,
                mic_volume=0.5,
                is_mic_mute=False,
            )
        raise InitializationActionError(action)

    if isinstance(action, SoundSetVolumeAction):
        if action.device == SoundDevice.OUTPUT:
            return replace(state, output_volume=action.volume)
        if action.device == SoundDevice.INPUT:
            return replace(state, input_volume=action.volume)
    elif isinstance(action, SoundChangeVolumeAction):
        if action.device == SoundDevice.OUTPUT:
            return replace(
                state,
                output_volume=min(
                    max(state.output_volume + action.amount, 0),
                    1,
                ),
            )
        if action.device == SoundDevice.INPUT:
            return replace(
                state,
                input_volume=min(
                    max(state.mic_volume + action.amount, 0),
                    1,
                ),
            )
    elif isinstance(action, SoundSetMuteStatusAction):
        if action.device == SoundDevice.OUTPUT:
            return replace(state, is_output_mute=action.mute)
        if action.device == SoundDevice.INPUT:
            return CompleteReducerResult(
                state=replace(state, is_mic_mute=action.mute),
                actions=[
                    StatusIconsRegisterAction(
                        icon='mic_off' if action.mute else 'mic',
                        priority=SOUND_MIC_STATE_ICON_PRIORITY,
                        id=SOUND_MIC_STATE_ICON_ID,
                    ),
                ],
            )
    elif isinstance(action, SoundToggleMuteStatusAction):
        return CompleteReducerResult(
            state=state,
            actions=[
                SoundSetMuteStatusAction(
                    mute=not state.is_output_mute
                    if action.device == SoundDevice.OUTPUT
                    else not state.is_mic_mute,
                    device=action.device,
                ),
            ],
        )
    return state



# ruff: noqa: D100, D101, D102, D103, D104, D107, N999
from __future__ import annotations

import pathlib
from typing import TYPE_CHECKING

from kivy.app import Builder
from ubo_gui.page import PageWidget

from ubo_app.store import dispatch
from ubo_app.store.camera import (
    CameraStartViewfinderAction,
)

if TYPE_CHECKING:
    from ubo_gui.menu.types import ActionItem

# Regular expression pattern
# WIFI:S:<SSID>;T:<WEP|WPA|blank>;P:<PASSWORD>;H:<true|false|blank>;;
barcode_pattern = r"""WIFI:S:(?P<SSID>[^;]*);(?:T:(?P<Type>(?i:WEP|WPA|WPA2|nopass));)\
?(?:P:(?P<Password>[^;]*);)?(?:H:(?P<Hidden>(?i:true|false));)?"""


class CreateWirelessConnectionPage(PageWidget):
    def get_item(self: CreateWirelessConnectionPage, index: int) -> ActionItem | None:
        if index == 2:  # noqa: PLR2004
            return {
                'label': 'start',
                'is_short': True,
                'icon': 'camera',
                'action': lambda: dispatch(
                    CameraStartViewfinderAction(barcode_pattern=barcode_pattern),
                ),
            }
        return super().get_item(index)


Builder.load_file(
    pathlib.Path(__file__)
    .parent.joinpath('create_wireless_connection.kv')
    .resolve()
    .as_posix(),
)



"""Implement `setup_service` function to set up and enable systemd service."""
import os
import subprocess
from pathlib import Path

from ubo_app.logging import logger

service_name = 'ubo_app'
service_content = """
[Install]
WantedBy=multi-user.target

[Unit]
Description=Ubo App Service
After=network.target

[Service]
User=root
ExecStart=/home/pi/ubo-app/bin/ubo
WorkingDirectory=/home/pi
StandardOutput=inherit
StandardError=inherit
Restart=always
Type=simple

[Install]
WantedBy=multi-user.target
"""


def setup_service(*, enable: bool = True, start: bool = True) -> None:
    """Create the service file and enable the service."""
    service_file_path = f'/etc/systemd/system/{service_name}.service'

    # Ensure we have the required permissions
    if os.geteuid() != 0:
        logger.error('This script needs to be run with root privileges.')
        return

    # Write the service content to the file
    with Path(service_file_path).open('w') as file:
        file.write(service_content)

    if enable:
        # Enable the service to start on boot
        subprocess.run(['/usr/bin/systemctl', 'enable', service_name], check=True)  # noqa: S603

    if start:
        # Start the service immediately
        subprocess.run(['/usr/bin/systemctl', 'start', service_name], check=True)  # noqa: S603

    logger.info(
        f"""Service '{service_name}' has been created and {
        'enabled' if enable else 'not enabled'}.""",
    )



# ruff: noqa: D100, D101, D102, D103, D104, D107
from __future__ import annotations

from dataclasses import field, replace
from typing import TYPE_CHECKING, Sequence, cast

from redux import (
    BaseAction,
    BaseEvent,
    CompleteReducerResult,
    FinishAction,
    Immutable,
    InitAction,
    InitializationActionError,
    ReducerResult,
)
from ubo_gui.menu.types import SubMenuItem

from ubo_app.store.app import (
    RegisterAppAction,
    RegisterRegularAppAction,
)
from ubo_app.store.keypad import (
    Key,
    KeypadAction,
    KeypadEvent,
    KeypadKeyPressAction,
    KeypadKeyPressEvent,
)
from ubo_app.store.sound import (
    SoundChangeVolumeAction,
    SoundDevice,
)
from ubo_app.store.status_icons import IconAction

if TYPE_CHECKING:
    from typing_extensions import TypeAlias
    from ubo_gui.menu.types import Menu


class MainState(Immutable):
    menu: Menu | None = None
    path: Sequence[int] = field(default_factory=list)


class InitEvent(BaseEvent):
    ...


class SetMenuPathAction(BaseAction):
    path: Sequence[str]


MainAction: TypeAlias = (
    InitAction
    | FinishAction
    | IconAction
    | KeypadAction
    | RegisterAppAction
    | SetMenuPathAction
)


def main_reducer(  # noqa: C901
    state: MainState | None,
    action: MainAction,
) -> ReducerResult[MainState, SoundChangeVolumeAction, KeypadEvent | InitEvent]:
    if state is None:
        if isinstance(action, InitAction):
            from ._menus import HOME_MENU

            return CompleteReducerResult(
                state=MainState(menu=HOME_MENU),
                events=[InitEvent()],
            )
        raise InitializationActionError(action)

    if isinstance(action, KeypadKeyPressAction):
        actions = []
        if action.key == Key.UP and len(state.path) == 0:
            actions.append(
                SoundChangeVolumeAction(
                    amount=0.05,
                    device=SoundDevice.OUTPUT,
                ),
            )
        if action.key == Key.DOWN and len(state.path) == 0:
            actions.append(
                SoundChangeVolumeAction(
                    amount=-0.05,
                    device=SoundDevice.OUTPUT,
                ),
            )
        return CompleteReducerResult(
            state=state,
            actions=actions,
            events=[
                KeypadKeyPressEvent(key=action.key),
            ],
        )

    if isinstance(action, RegisterAppAction):
        from ubo_gui.menu import Item, menu_items

        menu = state.menu
        parent_index = 0 if isinstance(action, RegisterRegularAppAction) else 1

        if not menu:
            return state

        root_menu_items = menu_items(menu)

        main_menu_item: Item = root_menu_items[0]
        if not isinstance(main_menu_item, SubMenuItem):
            msg = 'Main menu item is not a `SubMenuItem`'
            raise TypeError(msg)

        main_menu_items = menu_items(main_menu_item.sub_menu)

        desired_menu_item = main_menu_items[parent_index]
        if not isinstance(desired_menu_item, SubMenuItem):
            menu_title = (
                'Applications'
                if isinstance(action, RegisterRegularAppAction)
                else 'Settings'
            )
            msg = f'{menu_title} menu item is not a `SubMenuItem`'
            raise TypeError(msg)

        new_items = [
            *cast(Sequence[Item], desired_menu_item.sub_menu.items),
            action.menu_item,
        ]
        desired_menu_item = replace(
            desired_menu_item,
            sub_menu=replace(
                desired_menu_item.sub_menu,
                items=new_items,
            ),
        )
        main_menu_item = replace(
            main_menu_item,
            sub_menu=replace(
                main_menu_item.sub_menu,
                items=[
                    desired_menu_item if index == parent_index else item
                    for index, item in enumerate(main_menu_items)
                ],
            ),
        )

        return replace(
            state,
            menu=replace(
                menu,
                items=[
                    main_menu_item if index == 0 else item
                    for index, item in enumerate(root_menu_items)
                ],
            ),
        )

    if isinstance(action, SetMenuPathAction):
        return replace(state, path=action.path)

    return state



# ruff: noqa: D100, D101, D102, D103, D104, D107
from __future__ import annotations

from ubo_gui.app import UboApp

from .menu_central import MenuAppCentral
from .menu_footer import MenuAppFooter


class MenuApp(MenuAppCentral, MenuAppFooter, UboApp):
    """Menu application."""



# ruff: noqa: D100, D101, D102, D103, D104, D107, N999
from __future__ import annotations

from threading import current_thread
from typing import TYPE_CHECKING, Awaitable

from ubo_app.load_services import UboServiceThread
from ubo_app.logging import logger

if TYPE_CHECKING:
    from asyncio import Handle


background_tasks: set[Handle] = set()


def create_task(awaitable: Awaitable) -> Handle:
    async def wrapper() -> None:
        try:
            await awaitable
        except BaseException as exception:  # noqa: BLE001
            thread = current_thread()
            logger.exception(
                exception,
                extra={
                    'awaitable': awaitable,
                    **(
                        {'ubo_service': thread.path.as_posix()}
                        if isinstance(thread, UboServiceThread)
                        else {}
                    ),
                },
            )

    import _loop

    handle = _loop.create_task(wrapper())
    background_tasks.add(handle)
    return handle



# ruff: noqa: D100, D101, D102, D103, D104, D107
from __future__ import annotations

from functools import cached_property
from threading import Thread
from typing import TYPE_CHECKING

from kivy.clock import Clock
from kivy.metrics import dp
from kivy.uix.boxlayout import BoxLayout
from redux import EventSubscriptionOptions
from ubo_gui.app import UboApp
from ubo_gui.gauge import GaugeWidget
from ubo_gui.menu import MenuWidget
from ubo_gui.menu.constants import SHORT_WIDTH
from ubo_gui.page import PageWidget
from ubo_gui.volume import VolumeWidget

from ubo_app.store.keypad import Key, KeypadKeyPressEvent
from ubo_app.store.main import SetMenuPathAction

from .store import autorun, dispatch, subscribe_event

if TYPE_CHECKING:
    from kivy.uix.widget import Widget
    from ubo_gui.menu.types import Menu


class MenuAppCentral(UboApp):
    @cached_property
    def menu_widget(self: MenuAppCentral) -> MenuWidget:
        """Build the main menu and initiate it."""
        menu_widget = MenuWidget()

        @autorun(lambda state: state.main.current_menu)
        def sync_current_menu(selector_result: Menu | None) -> None:
            menu_widget.set_root_menu(selector_result)

        def handle_title_change(_: MenuWidget, title: str) -> None:
            self.root.title = title

        self.root.title = menu_widget.title
        menu_widget.bind(title=handle_title_change)
        menu_widget.bind(
            stack=lambda _, path: dispatch(
                SetMenuPathAction(
                    path=[
                        i.name if isinstance(i, PageWidget) else i[0]['title']
                        for i in path
                    ],
                ),
            ),
        )

        def handle_key_press_event(key_press_event: KeypadKeyPressEvent) -> None:
            if key_press_event.key == Key.L1:
                menu_widget.select(0)
            if key_press_event.key == Key.L2:
                menu_widget.select(1)
            if key_press_event.key == Key.L3:
                menu_widget.select(2)
            if key_press_event.key == Key.BACK:
                menu_widget.go_back()
            if key_press_event.key == Key.UP:
                menu_widget.go_up()
            if key_press_event.key == Key.DOWN:
                menu_widget.go_down()

        subscribe_event(
            KeypadKeyPressEvent,
            handle_key_press_event,
            EventSubscriptionOptions(run_async=False),
        )

        return menu_widget

    @cached_property
    def cpu_gauge(self: MenuAppCentral) -> GaugeWidget:
        import psutil

        gauge = GaugeWidget(value=0, fill_color='#24D636', label='CPU')

        value = 0

        def set_value(_: float) -> None:
            gauge.value = value

        def calculate_value() -> None:
            nonlocal value
            value = psutil.cpu_percent(interval=1, percpu=False)
            Clock.schedule_once(set_value)

        Clock.schedule_interval(
            lambda _: Thread(target=calculate_value).start(),
            1,
        )

        return gauge

    @cached_property
    def ram_gauge(self: MenuAppCentral) -> GaugeWidget:
        import psutil

        gauge = GaugeWidget(
            value=psutil.virtual_memory().percent,
            fill_color='#D68F24',
            label='RAM',
        )

        def set_value(_: int) -> None:
            gauge.value = psutil.virtual_memory().percent

        Clock.schedule_interval(set_value, 1)

        return gauge

    @cached_property
    def central(self: MenuAppCentral) -> Widget | None:
        horizontal_layout = BoxLayout()

        self.menu_widget.size_hint = (None, 1)
        self.menu_widget.width = dp(SHORT_WIDTH)
        horizontal_layout.add_widget(self.menu_widget)

        central_column = BoxLayout(
            orientation='vertical',
            spacing=dp(12),
            padding=dp(16),
        )
        central_column.add_widget(self.cpu_gauge)
        central_column.add_widget(self.ram_gauge)
        central_column.size_hint = (1, 1)
        horizontal_layout.add_widget(central_column)

        right_column = BoxLayout(orientation='vertical')
        volume_widget = VolumeWidget()
        right_column.add_widget(volume_widget)
        right_column.size_hint = (None, 1)
        right_column.width = dp(SHORT_WIDTH)
        horizontal_layout.add_widget(right_column)

        @autorun(
            lambda state: getattr(getattr(state, 'sound', None), 'output_volume', 0),
        )
        def sync_output_volume(selector_result: float) -> None:
            volume_widget.value = selector_result * 100
            self.root.render()

        def handle_depth_change(_: MenuWidget, depth: int) -> None:
            is_deep = depth > 0
            if is_deep:
                self.menu_widget.size_hint = (1, 1)
                central_column.size_hint = (0, 1)
                right_column.size_hint = (0, 1)
            else:
                self.menu_widget.size_hint = (None, 1)
                self.menu_widget.width = dp(SHORT_WIDTH)
                central_column.size_hint = (1, 1)
                right_column.size_hint = (None, 1)

        self.menu_widget.bind(depth=handle_depth_change)

        return horizontal_layout



# ruff: noqa: D100, D101, D102, D103, D104, D107
from __future__ import annotations

from typing import TYPE_CHECKING

from redux import BaseAction

if TYPE_CHECKING:
    from ubo_gui.menu import Item


class RegisterAppAction(BaseAction):
    menu_item: Item


class RegisterRegularAppAction(RegisterAppAction):
    ...


class RegisterSettingAppAction(RegisterAppAction):
    ...



from __future__ import annotations

from typing import TYPE_CHECKING, Sequence

from kivy.clock import Clock
from ubo_gui.menu.types import ActionItem, ApplicationItem, HeadlessMenu, SubMenuItem
from ubo_gui.notification import NotificationWidget

from ubo_app.logging import logger
from ubo_app.store import autorun, dispatch
from ubo_app.store.notifications import NotificationsClearAction

if TYPE_CHECKING:
    from ubo_gui.menu.types import Item

    from ubo_app.store.notifications import Notification

SETTINGS_MENU = HeadlessMenu(
    title='Settings',
    items=[],
)

APPS_MENU = HeadlessMenu(
    title='Apps',
    items=[],
)

MAIN_MENU = HeadlessMenu(
    title='Main',
    items=[
        SubMenuItem(
            label='Apps',
            icon='apps',
            sub_menu=APPS_MENU,
        ),
        SubMenuItem(
            label='Settings',
            icon='settings',
            sub_menu=SETTINGS_MENU,
        ),
        ActionItem(
            label='About',
            action=lambda: logger.info('"About" selected!'),
            icon='info',
        ),
    ],
)


@autorun(lambda store: store.notifications.unread_count)
def notifications_title(unread_count: int) -> str:
    return f'Notifications ({unread_count})'


@autorun(lambda store: store.notifications.notifications)
def notifications_menu_items(notifications: Sequence[Notification]) -> list[Item]:
    """Return a list of menu items for the notification manager."""

    def notification_widget_builder(
        notification: Notification,
        index: int,
    ) -> type[NotificationWidget]:
        """Return a notification widget for the given notification."""

        class NotificationWrapper(NotificationWidget):
            def __init__(self: NotificationWrapper, **kwargs: object) -> None:
                super().__init__(
                    **kwargs,
                    title=f'Notification ({index+1}/{len(notifications)})',
                    notification_title=notification.title,
                    content=notification.content,
                    color=notification.color,
                    icon=notification.icon,
                )

            def on_dismiss(self: NotificationWrapper) -> None:
                dispatch(
                    NotificationsClearAction(notification=notification),
                )
                Clock.schedule_once(lambda _: self.dispatch('on_close'), -1)

        return NotificationWrapper

    return [
        ApplicationItem(
            label=notification.title,
            icon=notification.icon,
            color='black',
            background_color=notification.color,
            application=notification_widget_builder(notification, index),
        )
        for index, notification in enumerate(notifications)
    ]


HOME_MENU = HeadlessMenu(
    title='Dashboard',
    items=[
        SubMenuItem(
            label='',
            sub_menu=MAIN_MENU,
            icon='menu',
            is_short=True,
        ),
        SubMenuItem(
            label='',
            sub_menu=HeadlessMenu(
                title=notifications_title,
                items=notifications_menu_items,
            ),
            color='yellow',
            icon='info',
            is_short=True,
        ),
        ActionItem(
            label='Turn off',
            action=lambda: logger.info('"Turn off" selected!'),
            icon='power_settings_new',
            is_short=True,
        ),
    ],
)



# ruff: noqa: D100, D101, D102, D103, D104, D107
from __future__ import annotations

from typing import TYPE_CHECKING, Callable

from kivy.base import Clock
from redux import (
    BaseCombineReducerState,
    CombineReducerAction,
    CreateStoreOptions,
    combine_reducers,
    create_store,
)

from ubo_app.logging import logger
from ubo_app.store.camera import CameraAction, CameraEvent
from ubo_app.store.ip import IpAction, IpEvent, IpState
from ubo_app.store.keypad import KeypadEvent
from ubo_app.store.main import MainAction, MainState, main_reducer
from ubo_app.store.notifications import NotificationsAction, NotificationsState
from ubo_app.store.sound import SoundAction, SoundState
from ubo_app.store.wifi import WiFiAction, WiFiEvent, WiFiState

from .status_icons import reducer as status_icons_reducer

if TYPE_CHECKING:
    from .status_icons import StatusIconsState


class RootState(BaseCombineReducerState):
    main: MainState
    status_icons: StatusIconsState
    sound: SoundState
    wifi: WiFiState
    ip: IpState
    notifications: NotificationsState


ActionType = (
    MainAction
    | CombineReducerAction
    | SoundAction
    | CameraAction
    | WiFiAction
    | IpAction
    | NotificationsAction
)
EventType = KeypadEvent | CameraEvent | WiFiEvent | IpEvent

root_reducer, root_reducer_id = combine_reducers(
    state_type=RootState,
    action_type=ActionType,
    event_type=EventType,
    main=main_reducer,
    status_icons=status_icons_reducer,
)


def scheduler(main_loop_callback: Callable[[], None]) -> None:
    Clock.create_trigger(
        lambda _: main_loop_callback(),
        0,
        interval=True,
    )()


store = create_store(
    root_reducer,
    CreateStoreOptions(
        auto_init=True,
        scheduler=scheduler,
        action_middleware=lambda action: logger.debug(
            'Action dispatched',
            extra={'action': action},
        ),
        event_middleware=lambda event: logger.debug(
            'Event dispatched',
            extra={'event': event},
        ),
    ),
)

autorun = store.autorun
dispatch = store.dispatch
subscribe = store.subscribe
subscribe_event = store.subscribe_event

__ALL__ = (autorun, dispatch, subscribe, subscribe_event)



# ruff: noqa: D100, D101, D102, D103, D104, D107, N999
from asyncio import Handle
from typing import Awaitable


def create_task(task: Awaitable) -> Handle:
    msg = f"Current thread is not an ubo service thread, can't run task {task}"
    raise NotImplementedError(msg)



# ruff: noqa: D100, D101, D102, D103, D104, D107, N999
from __future__ import annotations

from constants import SOUND_MIC_STATE_ICON_ID, SOUND_MIC_STATE_ICON_PRIORITY

from ubo_app.store import dispatch
from ubo_app.store.status_icons import StatusIconsRegisterAction


def init_service() -> None:
    dispatch(
        StatusIconsRegisterAction(
            icon='mic_off',
            priority=SOUND_MIC_STATE_ICON_PRIORITY,
            id=SOUND_MIC_STATE_ICON_ID,
        ),
    )
